import json 
from math import *
from decimal import *



#FUNCTIONS PARAMETERS
#--------------------------------------------
#peaks_detection()
current_min = 0.01                      # percentage of the max current #set the range of current in which look for maxima and minima...
current_max = 0.95                          # ... at the beginning and at the end of the procedure the current is in steady state ---> peaks are due to noise
Falling_window = 40                     # size of the windows in which determine if the central value is the local max/min ...
Rising_window = 5                           #... must be tuned depending on the switching speed of the valve and the sampling period
allowed_peaks_distance = 25             # minimum distance allowed between consecutive maxima if their value is 'too similar' (see function)

#---------------------------------------------
#current_standard_deviation()
Current_average_threshold = 0.0015      # milliAmps defining the range in which the current is considered in steady state at the end of the rising branch

#--------------------------------------------
# initial_inductance()
slope_window = 100                      # sets he amplitude of the window in which the average inductance is computed
Current_Threshold_Inductance = 0.03     # percentage of the max current # sets the initial value of current from which the computation of the average resistance starts
                                


################################################################################    
# prepares the structure of the new item for Dynamo DB

def Dynamo_DB_Psi_I(Rising, Falling):
    
    newItem = {}
    newItem["SwitchCounter"] = Rising["SwitchCount"]
    newItem["TestID"] = 1
    newItem["ValveID"] = Rising["valveID"]
    newItem["Time_stamp"] = Rising["timestamp"]
    newItem["Average_Resistance"] = Decimal(str(Rising["ResAve"]))
    newItem["Area"] = Decimal(str(- Psi_I_Area(Falling) - Psi_I_Area(Rising)))
   
   
    #______From the rising branch________________________________________________________________
    
    Flux = Rising["Flux"]
    Current = Rising["Current"]
    
    [ss_current, std_dev] = current_standard_deviation(Rising,Current_average_threshold)
    newItem["Steady_State_Current"] = Decimal(str(ss_current))
    newItem["Current_Standard_Deviation"] = Decimal(str(std_dev))
    
    [maxima, max_index, minima, min_index] = peaks_detection(Rising,Falling_window, Rising_window, std_dev)
    print("rising")
    print("max index: ", max_index)
    print("min index: ", min_index)
    print("switch count: ",Rising["SwitchCount"] )
    
    newItem["SW_ON_Start_Point_Current"] = Decimal(str(Current[max_index[0]])) 
    newItem["SW_ON_Start_Point_Flux"] = Decimal(str(Flux[max_index[0]]))
    
    newItem["SW_ON_End_Point_Current"] = Decimal(str(Current[min_index[0]]))
    newItem["SW_ON_End_Point_Flux"] = Decimal(str(Flux[min_index[0]]))
    
    newItem["Maximum_Point_Current"] = Decimal(str(Current[-1])) 
    newItem["Maximum_Point_Flux"] = Decimal(str(Flux[-1]))
    
    if(len(maxima)==len(minima)) and (len(maxima)>1):    
        newItem["Bounce_Max_Current"] = Decimal(str(Current[max_index[1]])) 
        newItem["Bounce_Max_Flux"] = Decimal(str(Flux[max_index[1]]))
        
        newItem["Bounce_Min_Current"] = Decimal(str(Current[min_index[1]]))
        newItem["Bounce_Min_Flux"] = Decimal(str(Flux[min_index[1]]))
        
    newItem["Delta_Curr_SW_ON"] = Decimal(str(-Current[max_index[0]] + Current[min_index[0]]))
    newItem["Delta_Flux_SW_ON"] = Decimal(str(-Flux[max_index[0]] + Flux[min_index[0]]))
    
    newItem["SW_ON_Path_Length"] = Decimal(str(Psi_I_path_length(max_index[0], min_index[0], Rising)))
    
    newItem["Switching_Time_ON"] = Decimal(str(Switching_Time(max_index[0], min_index[0], Rising["Tsampling"])))
    
    newItem["SW_ON_Time_Delay"] = Decimal(str(switching_delay(Rising, min_index[0])))
    
    newItem["Initial_Inductance"] = Decimal(str(initial_inductance(Flux, Current)))
    
    print("steady state current: ", ss_current)
    print("standard deviation: ", std_dev)
    

     #______From the falling branch_______________________________________________________________
    
    Flux = Falling["Flux"]
    Current = Falling["Current"]
    [maxima, max_index, minima, min_index] = peaks_detection(Falling, Falling_window, Rising_window, std_dev)
    print("falling")
    print("max index: ", max_index)
    print("min index: ", min_index)
    
    newItem["SW_OFF_Start_Point_Current"] =  Decimal(str(Current[min_index[0]])) 
    newItem["SW_OFF_Start_Point_Flux"] =  Decimal(str(Flux[min_index[0]]))
    newItem["SW_OFF_End_Point_Current"] = Decimal(str(Current[max_index[0]]))
    newItem["SW_OFF_End_Point_Flux"] = Decimal(str(Flux[max_index[0]]))
    
    newItem["Delta_Curr_SW_OFF"] = Decimal(str(Current[max_index[0]] - Current[min_index[0]]))
    newItem["Delta_Flux_SW_OFF"] = Decimal(str(Flux[max_index[0]] - Flux[min_index[0]]))
    
    newItem["SW_OFF_Path_Length"] = Decimal(str(Psi_I_path_length(max_index[0], min_index[0], Falling)))
    
    newItem["SW_OFF_Time_Delay"] = Decimal(str(switching_delay(Falling, max_index[0])))
    
    newItem["Switching_Time_OFF"] = Decimal(str(Switching_Time(max_index[0], min_index[0], Rising["Tsampling"])))
    
    
    return newItem




################################################################################    
# computes the area inside the switching cycle

def Psi_I_Area(content):
    
    A=0
    Flux = content["Flux"]
    Current = content["Current"]
    
    for i in range(0,len(Flux)-1):
        
        Dy = Flux[i+1] - Flux[i]
        y = Flux[i] + Dy/2
        Dx = Current[i+1] - Current[i]
        dA = y*Dx
        A += dA
        
    return A
    
    
    
################################################################################    
# detects the maxima and minima in a single branch

def peaks_detection(content, Falling_window, Rising_window, current_std_dev):
    
    Current = content["Current"]
    Flux=content["Flux"]
    Voltage = content["Voltage"]
    maxima = []
    max_index = []
    minima = []
    min_index = []
    
    if content["testID"] == 0:
        window_size = Falling_window#40
        position = 0
    else:
        window_size = Rising_window#4
        position = -1
        
    #________MAXIMA DETECTION_________________________________________
    
    for i in range(window_size , len(Current)-window_size):
        new_max = True
        
        if (Current[i] > current_min*Current[position]) & (Current[i]<current_max*Current[position]):
            for j in range (-window_size, window_size+1):
                if Current[i] < Current[i+j]:
                    new_max = False
                    break
            if (new_max == True):
                print("Max!", content["testID"])#kuagc
                #if (Current[i]- min(Current[i-window_size:i+window_size+1]) < 0.001):
                #    continue
                if len(maxima) != 0: #removes the previous maximum if it is too close to the new one in amplitude and position (--> almost flat part of the curve)
                    if (abs(Current[i] - maxima[-1]) < current_std_dev)&((i - max_index[-1])< allowed_peaks_distance):
                        print("Removed maximum!")
                        max_index.pop(-1)
                        maxima.pop(-1)
                else:
                    pass
                maxima.append(Current[i])
                max_index.append(i)
            else:
                pass
            
            
      #________MINIMA DETECTION___________________________________

    for i in range(window_size , len(Current)-window_size):
        new_min = True
        
        if (Current[i] > current_min*Current[position]) & (Current[i]<current_max*Current[position]):
            for j in range (-window_size, window_size+1):
                if Current[i] > Current[i+j]:
                    new_min = False
                    break
            if (new_min == True): # new minimum detected
                #if (max(Current[i-window_size : i+window_size+1])-Current[i]) < 0.001):
                #    continue
                print("Minimum!")
                if (content["testID"] == 1):
                    try:
                        if (i < max_index[len(min_index)]):
                            print("max after min")
                            continue
                    except:
                        print("exeption!")
                        pass # bounce not detected
                    #if (len(minima) == 0) & ((i - max_index[0])<60):
                    #    continue
                if len(minima) != 0:
                    if (Current[i] == minima[-1])&((i - min_index[-1])<window_size):
                        min_index.pop(-1)
                        minima.pop(-1)
                else:
                    pass
                minima.append(Current[i])
                min_index.append(i)
            else:
                pass
            
    return maxima, max_index, minima, min_index
    
    
    
################################################################################    
# compute the length of the path between 2 points

def Psi_I_path_length(point1, point2, Branch):
    
    Current = Branch["Current"]
    Flux = Branch["Flux"]
    
    length = 0
    if(Branch["testID"] == 1):
        for i in range(point1, point2 +1):
            length += sqrt( ((Flux[i+1]-Flux[i])**2) + ((Current[i+1] - Current[i])**2) )
    else:
        for i in range(point2, point1 +1):
            length += sqrt( ((Flux[i+1]-Flux[i])**2) + ((Current[i+1] - Current[i])**2) )
    
    return length
    


################################################################################    
# computes the average current and the standard deviation in steady state

def current_standard_deviation(rise_branch, Current_average_threshold):
    Current = rise_branch["Current"]
    length = len(Current)
    steady_state_current = 0
    
    for i in range(2,length):
        steady_state_current += Current[-i+1]
        if(Current[-1]-Current[-i]>Current_average_threshold):
            break
    steady_state_current = steady_state_current/i
    
    distances_sum = 0
    
    for j in range(1, i):
        distances_sum += sqrt( (steady_state_current - Current[length-i])**2 )
    
    standard_deviation = distances_sum/i
    
    return steady_state_current, standard_deviation
    
    
    
#################################################################################
# computes the amount of time among the instant in which input voltage is applied 
# and the switching is ended

def switching_delay(branch, end_switch_index):
    
    Voltage = branch["Voltage"]
    sampling_period =  branch["Tsampling"]
    
    
    if (branch["testID"]==1): # Rising branch
        for i in range(0,len(Voltage)):
            if(Voltage[i]>1):
                input_start_index = i
                break
        switching_delay = (end_switch_index - input_start_index) * sampling_period
        
    else:   # Falling branch
        for i in range(0,len(Voltage)):
            if(Voltage[i]<1):
                input_start_index = i
                break
        switching_delay = (end_switch_index - input_start_index) * sampling_period
        
    return switching_delay
    
    
#################################################################################
# computes the amount of time between the start and the end of the movement of the plunger.

def Switching_Time(start_index, end_index, sampling_period):
    
    samples_number = abs(start_index - end_index)
    
    return samples_number * sampling_period
    

#################################################################################
# computes the inductance of the solenoid at the beginning of the switchON branch
# (linear magnetic conditions and plunger doesn't move)

def initial_inductance(Flux, Current):
    
    for i in range(1, len(Current)):
        
        if (Current[i] > Current[-1]*Current_Threshold_Inductance):
            start = i
            break
        
    average_inductance = (Flux[i+slope_window]-Flux[i]) / (Current[i+slope_window] - Current[i])
    
    return average_inductance